/*
 * Copyright (c) 2008, Luc Yriarte
 * All rights reserved.
 * 
 * @file GSHHS.java
 * @author Luc Yriarte
 * 
 * License: BSD <http://www.opensource.org/licenses/bsd-license.php>
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted 
 * provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright notice, this 
 * list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright notice, this 
 * list of conditions and the following disclaimer in the documentation and/or other materials 
 * provided with the distribution.
 *     * The name of Luc Yriarte may not be used to endorse or promote products derived 
 * from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 */

package org.yriarte.carto;

import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Polygon;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;


/**
 * @author Luc Yriarte created on 1 sept. 2008 Paul Wessel's gshhs.c version
 *         available from Hawaii university site.
 * @see http://www.soest.hawaii.edu/pwessel/gshhs/gshhs.html
 */
public class GSHHS implements MapDataSource {

    private class gshhs_polygon {
        int id; /* Unique polygon id number, starting at 0 */

        int n; /* Number of points in this polygon */

        int flag; /* = level + version < < 8 + greenwich < < 16 + source < < 24 */

        /*
         * flag contains 4 items, one in each byte, as follows: low byte: level =
         * flag & 255: Values: 1 land, 2 lake, 3 island_in_lake, 4
         * pond_in_island_in_lake 2nd byte: version = (flag >> 8) & 255: Values:
         * Should be 4 for GSHHS version 1.4 3rd byte: greenwich = (flag >> 16) &
         * 255: Values: Greenwich is 1 if Greenwich is crossed 4th byte: source =
         * (flag >> 24) & 255: Values: 0 = CIA WDBII, 1 = WVS
         */
        int west, east, south, north; /* min/max extent in micro-degrees */

        int area; /* Area of polygon in 1/10 km^2 */

        int points[];

        public int level() {
            return flag & 0xff;
        }

        public int version() {
            return (flag >> 8) & 0xff;
        }

        public int greenwich() {
            return (flag >> 16) & 0xff;
        }

        public int src() {
            return (flag >> 24) & 0xff;
        }

    }

    private DataInputStream is;

    private Graphics graphics;

    private int x0, y0, width, height, west, east, south, north, px, py, dx, dy;


    public GSHHS() {
        super();
    }
    
    /**
     * @param width
     * @param height
     * @param west
     * @param east
     * @param south
     * @param north
     */
    public GSHHS(int x0, int y0,int width, int height, int west, int east, int south, int north) {
        super();
        init(x0, y0, width, height, west, east, south, north);
    }
    
    public void init(int x0, int y0,int width, int height, int west, int east, int south, int north) {
        this.x0 = x0;
        this.y0 = y0;
        this.width = width;
        this.height = height;
        this.west = west;
        this.east = east;
        this.south = south;
        this.north = north;
        /* micro degrees per pixel */
        px = (east - west) / width;
        py = (north - south) / height;
        dx = west + 180000000;
        dy = south + 90000000;
    }

    /**
     * @return Returns the graphics.
     */
    public Graphics getGraphics() {
        return graphics;
    }
    /**
     * @param graphics The graphics to set.
     */
    public void setGraphics(Graphics graphics) {
        this.graphics = graphics;
    }
    /**
     * @return Returns the InputStream.
     */
    public InputStream getInputStream() {
        return is;
    }
    /**
     * @param is The InputStream to set.
     */
    public void setInputStream(InputStream is) {
        this.is = new DataInputStream(is);
    }
    
    /* (non-Javadoc)
     * @see org.yriarte.carto.MapDataSource#drawMap(java.awt.Graphics, int, int, int, int, int, int, int, int)
     */
    public int drawMap(Graphics graphics, int x, int y, int width, int height,
            int west, int east, int south, int north) {
        // Method GSHHS.drawMap
        init(x0, y0, width, height, west, east, south, north);
        setGraphics(graphics);
        gshhs_polygon gshhs = null;
        while (true) {
            try {
                gshhs = readGshhsHeader();
            } catch (IOException e2) {
            	break;
            }
            if (gshhs == null)
                break;
            try {
                if ((gshhs.east-gshhs.west) / px > 2 && (gshhs.north-gshhs.south) / py > 2) {
                    Polygon poly = readAwtPolygon(gshhs);
                    graphics.drawPolyline(poly.xpoints,poly.ypoints,poly.npoints);
                    System.out.println(gshhs.id + " n:" + gshhs.n + " " + gshhs.level()
                            + " " + gshhs.src() + " area:" + gshhs.area + " w:" + gshhs.west
                            + " e:" + gshhs.east + " s:" + gshhs.south + " n:" + gshhs.north);
                }
                else {
                    is.skip(gshhs.n * 8);
//                    System.out.println(gshhs.id + " n:" + gshhs.n + " area:" + gshhs.area );
                }
            } catch (IOException e3) {
                // TODO Auto-generated catch block
                e3.printStackTrace();
            }
        }
        return 0;
    }
   /* 
    private Polygon readAwtPolygon() throws IOException {
        gshhs_polygon gshhs = readGshhsHeader();
        if (gshhs == null)
            return null;
        return readAwtPolygon(gshhs);
    }
*/
    private Polygon readAwtPolygon(gshhs_polygon gshhs) throws IOException {
        int xpoints[] = new int[gshhs.n];
        int ypoints[] = new int[gshhs.n];
        for (int i=0; i<gshhs.n; i++) {
            xpoints[i] = x0 + ( (dx + is.readInt()) % 360000000) / px;
            ypoints[i] = y0 - (dy + is.readInt()) / py + height / 2;
        }
        return new Polygon(xpoints, ypoints, gshhs.n);
    }

    gshhs_polygon readGshhsHeader() throws IOException {
        gshhs_polygon poly = new gshhs_polygon();
        poly.id = is.readInt();
        poly.n = is.readInt();
        poly.flag = is.readInt();
        poly.west = is.readInt();
        poly.east = is.readInt();
        poly.south = is.readInt();
        poly.north = is.readInt();
        poly.area = is.readInt();
        return poly;
    }

    public static void main(String[] args) {
        if (args.length != 1) {
            System.err.println("Parameter: gshhs binary data file.");
            System.exit(1);
        }

        GSHHS instance = new GSHHS();
        try {
            instance.setInputStream(new FileInputStream(args[0]));
        } catch (FileNotFoundException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            System.exit(1);
        }

        int width, heigth;
        width=720; heigth=360;
        Frame frame = new Frame();
        frame.setSize(width,heigth);
        frame.show();

        frame.getGraphics().drawLine(0,0,width,heigth);
        frame.getGraphics().drawLine(0,heigth,width,0);
        frame.getGraphics().drawLine(0,heigth/2,width,heigth/2);
        frame.getGraphics().drawLine(width/2,0,width/2,heigth);
//        instance.drawMap(frame.getGraphics(),0,0,width,heigth,0,360000000,-90000000,90000000);
//        instance.drawMap(frame.getGraphics(),0,0,width,heigth,-180000000,180000000,-90000000,90000000);
//        instance.drawMap(frame.getGraphics(),0,0,width,heigth,-120000000,240000000,-90000000,90000000);
      instance.drawMap(frame.getGraphics(),0,0,width,heigth,-180000000,180000000,-90000000,90000000);
        
        try {
            instance.getInputStream().close();
        } catch (IOException e1) {
            // 

            
            
            
            
            
            
            
            
            
            e1.printStackTrace();
            System.exit(1);
        }
    }

}
