/*
 * Copyright (c) 2008, Luc Yriarte
 * All rights reserved.
 * 
 * @file Matrix.java
 * @author Luc Yriarte
 * 
 * License: BSD <http://www.opensource.org/licenses/bsd-license.php>
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted 
 * provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright notice, this 
 * list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright notice, this 
 * list of conditions and the following disclaimer in the documentation and/or other materials 
 * provided with the distribution.
 *     * The name of Luc Yriarte may not be used to endorse or promote products derived 
 * from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 */

package org.yriarte.physicSim;

import org.yriarte.math.Vector;

/**
 * @brief PointMass : A weighted 3D Object 
 * Physics units in m / kg / s
 * Left handed geometric coordinates
 * 
 */
public class PointMass {
	/** By default objects weight 1 kg */
	public static double DEFAULT_MASS=1;
	/** mass in kg */
	protected double mass;
	protected double invmass;
	/** */
	protected Vector position;
	protected Vector velocity;
	protected Vector acceleration;
	
	/**
	 * 
	 */
	public PointMass() {
		super();
		this.mass = DEFAULT_MASS;
		this.position=new VectorXYZ();
		this.velocity=new VectorXYZ();
		this.acceleration=new VectorXYZ();
		this.invmass = 1/this.mass;
	}
	
	/**
	 * @param mass
	 */
	public PointMass(double mass) {
		super();
		this.mass = mass;
		this.position=new VectorXYZ();
		this.velocity=new VectorXYZ();
		this.acceleration=new VectorXYZ();
		this.invmass = 1/this.mass;
	}
	
	/**
	 * @brief Creates a point - mass with 0 speed and acceleration.
	 * @param mass in kg
	 * @param position a VectorXYZ in meters
	 */
	public PointMass(double mass, Vector position) {
		super();
		this.mass = mass;
		this.position = position;	
		this.velocity=new VectorXYZ();
		this.acceleration=new VectorXYZ();
		this.invmass = 1/this.mass;
//System.out.println("PointMass : mass=" + mass + " x=" + position.getCell()[0][0] 
//		+ " y=" + position.getCell()[1][0] 
//		+ " z=" + position.getCell()[2][0]);		
	}
	
	/**
	 * @return Returns the mass in kg.
	 */
	public double getMass() {
		return mass;
	}
	/**
	 * @param mass The mass in kg to set.
	 */
	public void setMass(double mass) {
		this.mass = mass;
		this.invmass = 1/this.mass;
	}
	
	public Vector getPosition() {
		return position;
	}
	
	/**
	 * @param position a VectorXYZ in meters
	 */
	public void setPosition(Vector position) {
		this.position = position;		
	}
	
	public Vector getVelocity() {
		return velocity;
	}
	
	public Vector getAcceleration() {
		return acceleration;
	}

	/**
	 * @param dt delta time in seconds
	 */
	public PointMass updateState(double dt) {
		velocity.addMulNumThis(acceleration, dt);
		position.addMulNumThis(velocity, dt);
		acceleration.toZero();
		return this;
	}
	
	/**
	 * @param f a VectorXYZ in m.kg/s
	 */
	public PointMass applyForce(Vector f) {
		acceleration.addMulNumThis(f, invmass);
		return this;
	}
	
}
