/*
 * Copyright (c) 2008, Luc Yriarte
 * All rights reserved.
 * 
 * @file Matrix.java
 * @author Luc Yriarte
 * 
 * License: BSD <http://www.opensource.org/licenses/bsd-license.php>
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted 
 * provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright notice, this 
 * list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright notice, this 
 * list of conditions and the following disclaimer in the documentation and/or other materials 
 * provided with the distribution.
 *     * The name of Luc Yriarte may not be used to endorse or promote products derived 
 * from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 */
package org.yriarte.physicSim;

import org.yriarte.math.Vector;
import org.yriarte.mini3D.Mesh;
import org.yriarte.mini3D.Vector3D;

/**
 * A spring on every edge of the mesh.
 */
public class SpringMesh extends Mesh {

	Spring springEdge[];
	PointMass springPoint[];
	
	/**
	 * 
	 */
	public SpringMesh() {
		super();
		springEdge = null;
		springPoint = null;
	}

	/**
	 * @param aMesh references the mesh vertices and edges
	 */
	public SpringMesh(Mesh aMesh) {
		super(aMesh);
		init();
	}

	/**
	 * @param nV
	 * @param nE
	 * @param v
	 * @param e
	 */
	public SpringMesh(int nV, int nE, Vector3D[] v, int[][] e) {
		super(nV, nE, v, e);
		init();
	}

	public SpringMesh init() {
		int i;
//System.out.println("SpringMesh init obj " + nVertex + " edge " + nEdge);
		springEdge = new Spring[nEdge];
		VectorXYZ v0,v1;
		for (i=0; i<nEdge; i++) {
			v0=VectorXYZ.MatrixAsVectorXYZ(vertex[edge[i][0]]);
			v1=VectorXYZ.MatrixAsVectorXYZ(vertex[edge[i][1]]);
			// Initial spring position is not stretched
			springEdge[i]=new Spring(new VectorXYZ(v0.getX() - v1.getX(),
					v0.getY() - v1.getY(),
					v0.getZ() - v1.getZ()));
		}
		springPoint = new PointMass[nVertex];
		// PointMass <x,y,z> and Vector3D <x,y,z,1> reference the same cells
		for (i=0; i<nVertex; i++) {
			springPoint[i] = new PointMass(1, VectorXYZ.MatrixAsVectorXYZ(vertex[i]));
		}
		return this;
	}
	
	public SpringMesh update(double dt) {
		int i;
//System.out.println("SpringMesh update dt " + dt + ", obj " + nVertex + " edge " + nEdge);
		// Calculate spring length
		Vector v0,v1;
		for (i=0; i<nEdge; i++) {
			v0=vertex[edge[i][0]];
			v1=vertex[edge[i][1]];
			springEdge[i].stretch = VectorXYZ.MatrixAsVectorXYZ(v0.sub(v1));
		}
		// Apply each spring force on edge's both ends 
		for (i=0; i<nEdge; i++) {
			springPoint[edge[i][0]].applyForce(Vector.MatrixAsVector(springEdge[i].getForceVector().mulNum(0.5 * dt)));
			springPoint[edge[i][1]].applyForce(Vector.MatrixAsVector(springEdge[i].getForceVector().mulNum(-0.5* dt)));
		}
		// Update <x,y,z> of edge's both ends Vector3D in place 
		for (i=0; i<nVertex; i++) {
			springPoint[i].updateState(dt);
		}
		return this;
	}

}
