/*
 * Copyright (c) 2008, Luc Yriarte
 * All rights reserved.
 * 
 * @file Matrix.java
 * @author Luc Yriarte
 * 
 * License: BSD <http://www.opensource.org/licenses/bsd-license.php>
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted 
 * provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright notice, this 
 * list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright notice, this 
 * list of conditions and the following disclaimer in the documentation and/or other materials 
 * provided with the distribution.
 *     * The name of Luc Yriarte may not be used to endorse or promote products derived 
 * from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 */
package org.yriarte.physicSim;

import java.awt.Component;

import org.yriarte.math.Vector;
import org.yriarte.mini3D.Buffered3DGraphics;
import org.yriarte.mini3D.Object3D;


public class World extends Thread {

	/** Time step in millisec for 12 fps */
	public static int DEFAULT_MS = 83;
	/** Time step in millisec */
	public int millisec = DEFAULT_MS; 

	/** By default gravity is 10 m.s-2 */
	public static double DEFAULT_GRAVITY=10;
	/** Gravity in this world */
	protected double g = DEFAULT_GRAVITY;
	protected VectorXYZ gravity;

	/** Objects in this world */
	protected SolidObject object[] = null;
	public int nObjects = 0;
	
	/** All solids are children of the space oject */
	public Object3D space;
	/** Current execution time in seconds */
	public double time=0;
	/** System time in milliseconds */
	public long systime;
	/** view graphics */
	public Buffered3DGraphics simView;
	/** applet */
	public Component view;
	
	/**
	 * 
	 */
	public World() {
		super();
		this.gravity=new VectorXYZ(0,-g,0);
		this.space = new Object3D();
	}

	/**
	 * @param gravity
	 */
	public World(double g) {
		super();
		this.g = g;
		this.gravity=new VectorXYZ(0,-g,0);
		this.space = new Object3D();
	}
	
	/**
	 * @param millisec
	 * @param g
	 */
	public World(double g, int millisec) {
		super();
		this.millisec = millisec;
		this.g = g;
		this.gravity=new VectorXYZ(0,-g,0);
		this.space = new Object3D();
}
	
	public int addObject(SolidObject obj) {
		int i;
		SolidObject[] prevObj = object;
		object = new SolidObject[nObjects + 1];
		for (i = 0; i < nObjects; i++) {
			object[i] = prevObj[i];
		}
		object[nObjects] = obj;
		space.addChild(obj.object3D);
		space.reset();
		nObjects++;
		return nObjects;
	}

	public SolidObject getObject(int i) {
		if (object == null)
			return null;
		return object[i];
	}
 
	/**
	 * @return Returns the g.
	 */
	public double getG() {
		return g;
	}
	
	/**
	 * @param g The g to set.
	 */
	public void setG(double g) {
		this.g = g;
		this.gravity=new VectorXYZ(0,-g,0);
	}
	
	public Buffered3DGraphics getSimView() {
		return simView;
	}
	
	public void setSimView(Buffered3DGraphics simView) {
		this.simView = simView;
		space.setDrawMode(simView.getDrawMode());		
	}
	
    public Object3D getSpace() {
        return space;
    }
    
    public void setSpace(Object3D space) {
        this.space = space;
    }
    
	public void updateObjects(double dt) {
		int i;
//		System.out.println("Update " + nObjects + " objects time=" + time + " dt=" + dt);
		if (g != 0) {
			Vector dg = new Vector(gravity.mulNum(dt));
			for (i = 0; i < nObjects; i++) {
				object[i].applyForce(dg);
			}
		}
		for (i = 0; i < nObjects; i++) {
			object[i].updateState(dt);
		}
	}
	
	public void updateWorld(double dt) {
		updateObjects(dt);
	}
	
	/**
	 * Implement Thread run method
	 */
	public void run() {
		double dt;
		time = 0;
		space.reset();
		while(true) {
			systime = System.currentTimeMillis();
			try {
				sleep(millisec);				
			} catch (Exception e) {
				System.out.println("runtime exception " + e.toString());
			}
			/* Calculate objects position velocity and acceleration */
			dt = (double) (System.currentTimeMillis() - systime) / 1000;
			updateWorld(dt);
			time+=dt;
			/* Redraw objects */
			simView.updateModel();
			view.repaint();
		}
	}

}
