/**
 * @author Luc Yriarte
 * 
 * Licence : BSD http://www.opensource.org/licenses/bsd-license.php
 *  
 */
package org.yriarte.test;

import java.applet.Applet;
import java.awt.Checkbox;
import java.awt.Choice;
import java.awt.Event;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.Observable;
import java.util.Observer;

import org.json.JSONException;
import org.json.JSONObject;
import org.json.XML;
import org.yriarte.awtGraphics.AwtGraphics;
import org.yriarte.awtGraphics.AwtImage;
import org.yriarte.mini3D.Buffered3DGraphics;
import org.yriarte.mini3D.Matrix3D;
import org.yriarte.mini3D.Object3D;
import org.yriarte.mini3D.PolyCylinder;
import org.yriarte.mini3D.PolySphere;
import org.yriarte.x3d.JSONx3dFactory;

public class X3dApplet extends Applet implements ActionListener, Observer, 
		ItemListener {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	final static int NAV_ROTATION = 0, NAV_TRANSLATION = 1, NAV_ZOOM = 2;
	final static int DEFAULT_FOCAL = 200, DEFAULT_Z = 400;

	int width, height;
	
	Checkbox drawFacets;
	Choice navMenu;

	Object3D model;
	Buffered3DGraphics bbG;
	int focal, dragX, dragY;
	double fz;

	public X3dApplet() {
		super();
		
		drawFacets = new Checkbox("Facets", true);
		drawFacets.addItemListener(this);

		navMenu = new Choice();
		navMenu.addItemListener(this);
		navMenu.addItem("Rotate");
		navMenu.addItem("Translate");
		navMenu.addItem("Zoom");
	}

	public void init() {
		width = this.getSize().width;
		height = this.getSize().height;
		String str;
		double z;
		
		str = getParameter("N");
		if (str != null)
			Object3D.N_DEFAULT = new Integer(str).intValue();
			
		str = getParameter("Z");
		z = (str == null) ? DEFAULT_Z : new Double(str).doubleValue();
		fz = z / ((width+height) / 2);
			
		str = getParameter("Focal");
		focal = (str == null) ? DEFAULT_FOCAL : new Integer(str).intValue();
			
		str = getParameter("Model");
		if (str == null) {
			System.err.println("No model, using sphere and cylinder");
			model = new PolySphere(100);
			model.addChild(new PolyCylinder(200,15));
			model.getChild(0).setPosition(Matrix3D.translation(115, 0, 0));
			model.getChild(0).setOrientation(Matrix3D.rotationX(45 * 2 * Math.PI / 360));
		}
		else {
			JSONObject jobj;
			try {
				jobj = XML.toJSONObject(str);
				model = JSONx3dFactory.getScene(jobj);
			} catch (JSONException e) {
				System.err.println("x3d model error: " + e);
				model = new PolySphere(100);
			}
			
		}

		str = getParameter("Nav");
		if (str == null)
			add(navMenu);
		else if (str.equalsIgnoreCase("Rotate"))
			navMenu.select(NAV_ROTATION);
		else if (str.equalsIgnoreCase("Translate"))
			navMenu.select(NAV_TRANSLATION);
		else if (str.equalsIgnoreCase("Zoom"))
			navMenu.select(NAV_ZOOM);
		
		str = getParameter("Facets");
		if (str == null)
			add(drawFacets);
		else drawFacets.setState(str.equalsIgnoreCase("true"));
		
		bbG = new Buffered3DGraphics(model, new AwtImage(this.createImage(width,height)), width, height, focal);
		bbG.setDrawMode(drawFacets.getState() ? Object3D.DRAW_FACETS : Object3D.DRAW_WIREFRAME);
		
		model.setPosition(Matrix3D.translation(0, 0, z));
		model.reset();
		
		// ---- ---- update and paint
		bbG.updateModel();
		repaint();
	}
	
	public void paint(Graphics g) {
		bbG.updateView(new AwtGraphics(g));
	}

	public void update(Graphics g) {
		this.paint(g);
	}

	public boolean mouseDown(Event arg0, int arg1, int arg2) {
		dragX = arg1;
		dragY = arg2;
		return true;
	}

	public boolean mouseDrag(Event arg0, int arg1, int arg2) {
		double deltaX, deltaY, deltaZ;
		if (navMenu.getSelectedIndex() == NAV_ROTATION) {
			deltaX = (arg1 - dragX) * 2 * Math.PI / width;
			deltaY = (dragY - arg2) * 2 * Math.PI / height;
			model.setOrientation(model.getOrientation().mul(
							Matrix3D.rotationY(deltaX).mul(
									Matrix3D.rotationX(deltaY))));
		} else if (navMenu.getSelectedIndex() == NAV_TRANSLATION) {
			deltaX = fz * (arg1 - dragX);
			deltaY = fz * (dragY - arg2);
			model.setPosition(model.getPosition().mul(
					Matrix3D.translation(deltaX, deltaY, 0)));
		} else if (navMenu.getSelectedIndex() == NAV_ZOOM) {
			deltaZ = fz * (arg2 - dragY);
			model.setPosition(model.getPosition().mul(
					Matrix3D.translation(0, 0, deltaZ)));
		}
		bbG.updateModel();
		repaint();
		dragX = arg1;
		dragY = arg2;
		return true;
	}

	public boolean mouseUp(Event arg0, int arg1, int arg2) {
		dragX = dragY = 0;
		return true;
	}

	public void itemStateChanged(ItemEvent arg0) {
		if (arg0.getSource() == drawFacets) {
			bbG.setDrawMode(drawFacets.getState() ? Object3D.DRAW_FACETS : Object3D.DRAW_WIREFRAME);
			bbG.updateModel();
			repaint();
		}
	}

	public void actionPerformed(ActionEvent arg0) {
//		Object src = arg0.getSource();
	}

	public void update(Observable arg0, Object arg1) {
		repaint();
	}

}
