/**
 * @author Luc Yriarte
 * 
 * Licence : BSD http://www.opensource.org/licenses/bsd-license.php
 *  
 */
package org.yriarte.wiki3D;

import java.applet.Applet;
import java.awt.Button;
import java.awt.Checkbox;
import java.awt.Choice;
import java.awt.Event;
import java.awt.Graphics;
import java.awt.TextArea;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.Observable;
import java.util.Observer;

import org.yriarte.awtGraphics.AwtGraphics;
import org.yriarte.awtGraphics.AwtImage;
import org.yriarte.mini3D.Animation;
import org.yriarte.mini3D.Buffered3DGraphics;
import org.yriarte.mini3D.Matrix3D;
import org.yriarte.mini3D.Object3D;
import org.yriarte.mini3D.PolySphere;
import org.yriarte.mini3D.Vector3D;

public class Wiki3DApplet extends Applet implements ActionListener, Observer,
		ItemListener {

	static int NAV_ROTATION = 0, NAV_TRANSLATION = 1, NAV_ZOOM = 2, NAV_SELECT = 3;
	static int DEFAULT_S = 10, DEFAULT_FOCAL = 200, DEFAULT_Z = 200, DEFAULT_MS = 83; // 12 fps

	int width, height;
	
	Checkbox drawFacets;
	Checkbox showText;
	Choice navMenu;
	TextArea codeField;
	Button readButton;

	Object3D model;
	Animation anime;
	Buffered3DGraphics graphics3D;
	int ms, focal, dragX, dragY;

	public Wiki3DApplet() {
		super();
		anime = null;
		
		drawFacets = new Checkbox("Facets", true);
		drawFacets.addItemListener(this);

		navMenu = new Choice();
		navMenu.addItemListener(this);
		navMenu.addItem("Rotate");
		navMenu.addItem("Translate");
		navMenu.addItem("Zoom");
		navMenu.addItem("Select");

		showText = new Checkbox("Show text");
		showText.addItemListener(this);
	}

	public void showTextControls() {
		readButton = new Button("Create model");
		add(readButton);
		readButton.addActionListener(this);
		codeField = new TextArea(5, 30);
		add(codeField);
		codeField.setText(model.toString());
		validate();
	}

	public void hideTextControls() {
		remove(readButton);
		remove(codeField);
		readButton = null;
		codeField = null;
		validate();
	}

	public void init() {
		width = this.getSize().width;
		height = this.getSize().height;
		String str;
		double z;
		
		str = getParameter("N");
		if (str != null)
			Object3D.N_DEFAULT = new Integer(str).intValue();
			
		str = getParameter("Z");
		z = (str == null) ? DEFAULT_Z : new Double(str).doubleValue();
			
		str = getParameter("Focal");
		focal = (str == null) ? DEFAULT_FOCAL : new Integer(str).intValue();
			
		str = getParameter("ms");
		ms = (str == null) ? DEFAULT_MS : new Integer(str).intValue();
			
		str = getParameter("Model");
		model = (str == null) ? new PolySphere(100) : Object3DFactory.NewObject3D(str);

		str = getParameter("Edit");
		if (str != null && str.equals("true")) add(showText);
		
		str = getParameter("Nav");
		if (str == null)
			add(navMenu);
		else if (str.equals("Rotate"))
			navMenu.select(NAV_ROTATION);
		else if (str.equals("Translate"))
			navMenu.select(NAV_TRANSLATION);
		else if (str.equals("Zoom"))
			navMenu.select(NAV_ZOOM);
		else if (str.equals("Select"))
			navMenu.select(NAV_SELECT);
		
		str = getParameter("Facets");
		if (str == null)
			add(drawFacets);
		else if (str.equals("false"))
			drawFacets.setState(false);
		
		graphics3D = new Buffered3DGraphics(model, new AwtImage(this.createImage(width,height)), width, height, focal);
		graphics3D.setDrawMode(drawFacets.getState() ? Object3D.DRAW_FACETS : Object3D.DRAW_WIREFRAME);
		
		model.setPosition(Matrix3D.translation(0, 0, z));
		
		// ---- ---- update and paint
		graphics3D.updateModel();
		repaint();
		
		// ---- ---- parsed animation ?
		if (Object3DFactory.getAnimation() != null) {
			anime = Object3DFactory.getAnimation();
			Object3DFactory.resetAnimation();
			anime.addObserver(this);
			anime.setGraphics(graphics3D);
			Thread animeThread = new Thread(anime);
			animeThread.start();
		}		
	}
	
	public void paint(Graphics g) {
		graphics3D.updateView(new AwtGraphics(g));
	}

	public void update(Graphics g) {
		this.paint(g);
	}

	public boolean mouseDown(Event arg0, int arg1, int arg2) {
		dragX = arg1;
		dragY = arg2;
		if (navMenu.getSelectedIndex() == NAV_SELECT)
			model.selectVertex(arg1,arg2,DEFAULT_S);
		return true;
	}

	public boolean mouseDrag(Event arg0, int arg1, int arg2) {
		double deltaX, deltaY, deltaZ;
		if (navMenu.getSelectedIndex() == NAV_ROTATION) {
			deltaX = (arg1 - dragX) * 2 * Math.PI / width;
			deltaY = (dragY - arg2) * 2 * Math.PI / height;
			model.setOrientation(model.getOrientation().mul(
							Matrix3D.rotationY(deltaX).mul(
									Matrix3D.rotationX(deltaY))));
		} else if (navMenu.getSelectedIndex() == NAV_TRANSLATION) {
			deltaX = (arg1 - dragX);
			deltaY = (dragY - arg2);
			model.setPosition(model.getPosition().mul(
					Matrix3D.translation(deltaX, deltaY, 0)));
		} else if (navMenu.getSelectedIndex() == NAV_ZOOM) {
			deltaZ = arg2 - dragY;
			model.setPosition(model.getPosition().mul(
					Matrix3D.translation(0, 0, deltaZ)));
		} else if (navMenu.getSelectedIndex() == NAV_SELECT) {
			deltaX = (arg1 - dragX);
			deltaY = (dragY - arg2);
			Vector3D v = model.getSelectedVertex();
			if (v != null) {
				v.setX(v.getX()+deltaX);
				v.setY(v.getY()+deltaY);
			}
		}
		graphics3D.updateModel();
		repaint();
		dragX = arg1;
		dragY = arg2;
		return true;
	}

	public boolean mouseUp(Event arg0, int arg1, int arg2) {
		dragX = dragY = 0;
		return true;
	}

	public void itemStateChanged(ItemEvent arg0) {
		if (arg0.getSource() == showText) {
			if (showText.getState())
				showTextControls();
			else
				hideTextControls();
		} else if (arg0.getSource() == drawFacets) {
			graphics3D.setDrawMode(drawFacets.getState() ? Object3D.DRAW_FACETS : Object3D.DRAW_WIREFRAME);
			graphics3D.updateModel();
			repaint();
		}
	}

	public void actionPerformed(ActionEvent arg0) {
//		Object src = arg0.getSource();
	}

	public void update(Observable arg0, Object arg1) {
		repaint();
	}


}
