/*
 * Copyright (c) 2008, Luc Yriarte
 * All rights reserved.
 * 
 * @file JSONx3dFactory.java
 * @author Luc Yriarte
 * 
 * License: BSD <http://www.opensource.org/licenses/bsd-license.php>
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted 
 * provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright notice, this 
 * list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright notice, this 
 * list of conditions and the following disclaimer in the documentation and/or other materials 
 * provided with the distribution.
 *     * The name of Luc Yriarte may not be used to endorse or promote products derived 
 * from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 * 
 */
package org.yriarte.x3d;

import java.util.HashMap;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.yriarte.graphics.Color;
import org.yriarte.mini3D.Box;
import org.yriarte.mini3D.Matrix3D;
import org.yriarte.mini3D.Mesh;
import org.yriarte.mini3D.Object3D;
import org.yriarte.mini3D.PolyCone;
import org.yriarte.mini3D.PolyCylinder;
import org.yriarte.mini3D.PolySphere;
import org.yriarte.mini3D.Vector3D;
import org.yriarte.mini3D.Object3D.PolygonEdgeIndexes;

public class JSONx3dFactory {

	static HashMap defs = new HashMap();
	
	static void defKey(JSONObject jObj) {
		try {
			String key = jObj.getString("DEF");
//System.out.println("--- DEF " + key + " ---> ");
//System.out.println(jObj + "\n");
			defs.put(key,jObj);
		} catch (JSONException e) {
		}
	}
	
	static JSONObject useKey(JSONObject jObj) {
		JSONObject jUse = null;
		try {
			String key = jObj.getString("USE");
		 	jUse = (JSONObject) defs.get(key);
		} catch (JSONException e) {
		}		
		return jUse == null ? jObj : jUse;
	}
	
	/**
	 * @param aTransformGroup a "Transform" or a "Group" node
	 * @return an Object3D
	 */
	static Object3D getObject3D(JSONObject aTransformGroup) {
		defKey(aTransformGroup);
		aTransformGroup = useKey(aTransformGroup);
		Object3D obj = null;
		try {
			JSONObject aShape = aTransformGroup.getJSONObject("Shape");
			obj = getShape(aShape);
			obj.setColor(getColor(aShape));
		} catch (JSONException e) {
			try {
				JSONArray jShapes = aTransformGroup.getJSONArray("Shape");
				obj = getShape(jShapes.getJSONObject(0));
				obj.setColor(getColor(jShapes.getJSONObject(0)));
				for (int i=1; i<jShapes.length(); i++) {
					Object3D child = getShape(jShapes.getJSONObject(i));
					child.setColor(getColor(jShapes.getJSONObject(i)));
					obj.addChild(child);
				}
			} catch (JSONException e1) {
				obj = new Object3D();
			}
		}
		obj.setPosition(getTranslation(aTransformGroup));
		obj.setOrientation(getRotation(aTransformGroup));
		Matrix3D scale = getScale(aTransformGroup);
		String[] tags = {"Transform","Group"};
		for (int s=0; s<tags.length;s++) {
			try {
				JSONArray jObjects = aTransformGroup.getJSONArray(tags[s]);
				for (int i=0; i<jObjects.length(); i++)
					obj.addChild(getObject3D(jObjects.getJSONObject(i)));
			} catch (JSONException e) { 
				try {
					obj.addChild(getObject3D(aTransformGroup.getJSONObject(tags[s])));
				} catch (JSONException e1) {
				}
			}
		}
		if (!scale.isId())
		    obj.transformMeshes(scale);
		return obj;
	}
	
	/**
	 * @param aShape a "Shape" node
	 * @return an Object3D. Can be a "Cone", a "Cylinder", a "Sphere", a "Box" or an "IndexedFaceSet"
	 */
	static Object3D getShape(JSONObject aShape) {
		defKey(aShape);
		aShape = useKey(aShape);
		Object3D obj = null;
		try {
			aShape.getJSONObject("Cone");
			double h,r;
			try {
				h = aShape.getJSONObject("Cone").getDouble("height");
			} catch (JSONException e) {h=1;}
			try {
				r = aShape.getJSONObject("Cone").getDouble("bottomRadius");
			} catch (JSONException e) {r=1;}
			obj = new PolyCone(h,r);
		} catch (JSONException eCone) { try {
			aShape.getJSONObject("Cylinder");
			double h,r;
			try {
				h = aShape.getJSONObject("Cylinder").getDouble("height");
			} catch (JSONException e) {h=1;}
			try {
				r = aShape.getJSONObject("Cylinder").getDouble("radius");
			} catch (JSONException e) {r=1;}
			obj = new PolyCylinder(h,r);
		} catch (JSONException eCylinder) { try {
			aShape.getJSONObject("Sphere");
			double r;
			try {
				r = aShape.getJSONObject("Sphere").getDouble("radius");
			} catch (JSONException e) {r=1;}
			obj = new PolySphere(r);
		} catch (JSONException eSphere) { try {
			aShape.getJSONObject("Box");
			double x,y,z;
			try {
				String xyz[] = aShape.getJSONObject("Box").getString("size").split("[ \t\n]");
				x=Double.parseDouble(xyz[0]);
				y=Double.parseDouble(xyz[1]);
				z=Double.parseDouble(xyz[2]);
			} catch (Exception e) {x=y=z=1;}
			obj = new Box(x,y,z);
		} catch (JSONException eFaceSet) { try {
		    int i,j,k,l,m;
		    JSONObject indexedFaceSet = aShape.getJSONObject("IndexedFaceSet");
			String indexstr = indexedFaceSet.getString("coordIndex").trim().replaceAll(",","");
			String indexes[] = indexstr.split("[ \t\n]");
			String pointstr = indexedFaceSet.getJSONObject("Coordinate").getString("point").trim().replaceAll(",","");
			String xyzPoints[] = pointstr.split("[ \t\n]");
			int nPoints = xyzPoints.length/3;
//System.out.println("====> " + nPoints + "/" + xyzPoints.length + " points: " + pointstr);			
			Vector3D vertex[] = new Vector3D[nPoints];
			for (i=0; i<nPoints; i++) {
//System.out.println("[" + xyzPoints[i*3] + "," + xyzPoints[i*3+1] + "," + xyzPoints[i*3+2] + "]");			
			    vertex[i] = new Vector3D(Double.parseDouble(xyzPoints[i*3]),Double.parseDouble(xyzPoints[i*3+1]),Double.parseDouble(xyzPoints[i*3+2]));
			}
			i=j=0;
			while(i != -1) {
			    i=indexstr.indexOf("-1",i+1);
			    j++;
			}
			int nFaces = j-1;
			int nEdges = indexes.length-nFaces;
//System.out.println("====> " + indexes.length + " indexes, " + nFaces + " faces, " + nEdges + " edges: " + indexstr);			
			int edges[][] = new int[nEdges][2];
			PolygonEdgeIndexes faces[] = new PolygonEdgeIndexes[nFaces];
			obj = new Object3D();
			j=k=l=m=0;
			for (i=0; i<indexes.length; i++) {
//System.out.println("i=" + i + "(" + indexes[i] + ")");			
			    if (Integer.parseInt(indexes[i]) == -1) {
		            faces[m] = obj.new PolygonEdgeIndexes(obj, i-j);
			        for (k=j; k<i; k++) {
			            faces[m].edge[k-j]=l;
//System.out.println("l=" + l + " k=" + k + " j=" + j + " i=" + i + "(" + indexes[k] + "," + indexes[k+1] + ")");
			            edges[l][0]=Integer.parseInt(indexes[k]);
			            edges[l][1]=Integer.parseInt(indexes[k+1]);
			            l++;
			        }
			        m++;
//System.out.println("(" + edges[l-1][0] + "," + edges[l-1][1] + ") => (" + edges[l-1][0] + "," + indexes[j] + ")");			
		            edges[l-1][1]=Integer.parseInt(indexes[j]);
			        j=i+1;
			    }
			}
			obj.setMesh(new Mesh(vertex.length,edges.length,vertex,edges));
			obj.setFacet(faces,nFaces);
		} catch (JSONException e) {
System.err.println("Unknown shape: " + aShape.toString());			
			obj = new Object3D();
		}}}}	}
		return obj;
	}
	
	/**
	 * @param aShape a "Shape" node.
	 * @return a Color
	 */
	static Color getColor(JSONObject aShape) {
		Color color = null;
		try {
			JSONObject appear = aShape.getJSONObject("Appearance");
			defKey(appear);
			appear = useKey(appear);
			String rgb[] = appear.getJSONObject("Material").getString("diffuseColor").split("[ \t\n]");
			color = new Color(Float.parseFloat(rgb[0]),Float.parseFloat(rgb[1]),Float.parseFloat(rgb[2]));
		} catch (Exception e) {
		}
		return color;
	}
	
	/**
	 * @param aTransform a "Transform" node
	 * @return a translation matrix
	 */
	static Matrix3D getTranslation(JSONObject aTransform) {
	    return getTranslationMatrix(aTransform,"translation");
	}
	static Matrix3D getPosition(JSONObject aTransform) {
		Matrix3D m3d = null;
		try {
		    aTransform.getJSONObject("position");
			m3d = getTranslationMatrix(aTransform,"position");   		    
		} catch (JSONException e) {
			m3d = Matrix3D.translation(0,0,-10);
		}		
//System.out.println("position     : " + m3d);
//System.out.println("position -1  : " + Matrix3D.MatrixAsMatrix3D(m3d.mulNum(-1)));
//System.out.println("position inv : " + Matrix3D.MatrixAsMatrix3D(m3d.inv()));
		return Matrix3D.MatrixAsMatrix3D(m3d.inv());
	}
	static Matrix3D getTranslationMatrix(JSONObject aTransform, String str) {
		Matrix3D m3d = null;
		try {
			String xyz[] = aTransform.getString(str).split("[ \t\n]");
			m3d = Matrix3D.translation(Double.parseDouble(xyz[0]),Double.parseDouble(xyz[1]),-Double.parseDouble(xyz[2]));
		} catch (JSONException e) {
			m3d = new Matrix3D();
		}		
		return m3d;
	}
	
	/**
	 * @param aTransform a "Transform" node
	 * @return a rotation matrix
	 */
	static Matrix3D getRotation(JSONObject aTransform) {
	    return getRotationMatrix(aTransform,"rotation");
	}
	static Matrix3D getOrientation(JSONObject aTransform) {
//System.out.println("orientation     : " + getRotationMatrix(aTransform,"orientation"));
//System.out.println("orientation -1  : " + Matrix3D.MatrixAsMatrix3D(getRotationMatrix(aTransform,"orientation").mulNum(-1)));
//System.out.println("orientation inv : " + Matrix3D.MatrixAsMatrix3D(getRotationMatrix(aTransform,"orientation").inv()));
	    return Matrix3D.MatrixAsMatrix3D(getRotationMatrix(aTransform,"orientation").inv());
	}
	static Matrix3D getRotationMatrix(JSONObject aTransform, String str) {
		Matrix3D m3d = new Matrix3D();
		try {
			String xyza[] = aTransform.getString(str).split("[ \t\n]+");
			double fx = Double.parseDouble(xyza[0]);
			double fy = Double.parseDouble(xyza[1]);
			double fz = Double.parseDouble(xyza[2]);
			double teta = Double.parseDouble(xyza[3]);
			if (fx != 0)
				m3d = Matrix3D.MatrixAsMatrix3D(m3d.mul(Matrix3D.rotationX(teta * fx)));
			if (fy != 0)
				m3d = Matrix3D.MatrixAsMatrix3D(m3d.mul(Matrix3D.rotationY(teta * fy)));
			if (fz != 0)
				m3d = Matrix3D.MatrixAsMatrix3D(m3d.mul(Matrix3D.rotationZ(teta * fz)));
		} catch (JSONException e) {
			m3d = new Matrix3D();
		}		
		return m3d;
	}
	
	/**
	 * @param aTransform a "Transform" node
	 * @return a scale matrix
	 */
	static Matrix3D getScale(JSONObject aTransform) {
	    return getScaleMatrix(aTransform,"scale");
	}
	static Matrix3D getScaleMatrix(JSONObject aTransform, String str) {
		Matrix3D m3d = null;
		try {
			String xyz[] = aTransform.getString(str).split("[ \t\n]");
			m3d = Matrix3D.scale(Double.parseDouble(xyz[0]),Double.parseDouble(xyz[1]),-Double.parseDouble(xyz[2]));
		} catch (JSONException e) {
			m3d = new Matrix3D();
		}
		return m3d;
	}
	
	
	/**
	 * @param toplevel x3d toplevel node 
	 * @return an Object3D for the toplevel scene
	 */
	static public Object3D getScene(JSONObject toplevel) {
		JSONObject jScene = null;
		try {
			jScene = toplevel.getJSONObject("X3D").getJSONObject("Scene").getJSONObject("ProtoDeclare").getJSONObject("ProtoBody");
		} catch (Exception e) {
			try {
				jScene = toplevel.getJSONObject("X3D").getJSONObject("Scene");
			} catch (Exception e1) {
				return new Object3D();
			}
		}
		Object3D scene = getObject3D(jScene);
		scene.setPosition(getPosition(jScene));
		scene.setOrientation(getOrientation(jScene));
		scene.reset();
		return scene;
	}
	
}


	
